@@ -13,9 +13,16 @@ log = Logger("opengl")
 required_extensions = ["GL_ARB_texture_rectangle", "GL_ARB_vertex_program"]
 
 
-BLACKLIST = {"vendor" : ["nouveau", "Humper", "VMware, Inc."]}
-if sys.platform.startswith("darwin"):
-    #crashes were reported with the Intel driver on OSX
+WHITELIST = {}
+BLACKLIST = {"vendor" : ["nouveau", "Humper", "VMware, Inc."],
+             "renderer" : ["Software Rasterizer"]}
+if False:
+    #for testing:
+    BLACKLIST["vendor"].append("NVIDIA Corporation")
+    WHITELIST["renderer"] = ["GeForce GTX 760/PCIe/SSE2"]
+    
+if sys.platform.startswith("darwin") or sys.platform.startswith("win"):
+    #crashes were reported with the Intel driver on OSX 
     BLACKLIST["vendor"].append("Intel Inc.")
 
 
@@ -91,6 +98,31 @@ def raise_error(msg):
 gl_check_error = raise_error
 
 
+#support for memory views requires Python 2.7 and PyOpenGL 3.1
+def is_pyopengl_memoryview_safe(pyopengl_version):
+    if sys.version_info[:2]<(2,7):
+        #memoryview type requires python 2.7
+        return False
+    vsplit = pyopengl_version.split('.')
+    if vsplit[:2]<['3','1']:
+        #requires PyOpenGL >= 3.1, earlier versions will not work
+        return False
+    if vsplit[:2]>=['3','2']:
+        #assume that newer versions are OK too
+        return True
+    #at this point, we know we have a 3.1.x version, but which one?
+    if len(vsplit)<3:
+        #not enough parts to know for sure, assume it's not supported
+        return False
+    micro = vsplit[2]
+    #ie: '0', '1' or '0b2'
+    if micro=='0':
+        return True     #3.1.0 is OK
+    if micro>='1':
+        return True     #3.1.1 onwards should be too
+    return False        #probably something like '0b2' which is broken
+
+
 def check_functions(*functions):
     missing = []
     available = []
@@ -146,9 +178,18 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
                               (".".join([str(x) for x in MIN_VERSION]), gl_major, gl_minor))
         else:
             log("found valid OpenGL version: %s.%s", gl_major, gl_minor)
+
+	from OpenGL import version as OpenGL_version
+        try:
+            import OpenGL_accelerate
+        except:
+            OpenGL_accelerate = None
+        props["zerocopy"] = is_pyopengl_memoryview_safe(OpenGL_version.__version__) and bool(OpenGL_accelerate)
+
         try:
             extensions = glGetString(GL_EXTENSIONS).split(" ")
         except:
+            extensions = []
             gl_check_error("OpenGL could not find the list of GL extensions - does the graphics driver support OpenGL?")
         log("OpenGL extensions found: %s", ", ".join(extensions))
         props["extensions"] = extensions
@@ -170,7 +211,7 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
                 if fatal:
                     gl_check_error("OpenGL property '%s' is missing" % d)
                 else:
-                    log.warn("OpenGL property '%s' is missing" % d)
+                    log("OpenGL property '%s' is missing", d)
                 v = ""
             props[d] = v
 
@@ -180,13 +221,25 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
             log("%s: %s", d, v)
             props[d] = v
 
+        blacklisted = None
+        whitelisted = None
         for k,vlist in BLACKLIST.items():
             v = props.get(k)
             if v in vlist:
-                if force_enable:
-                    log.warn("Warning: %s '%s' is blacklisted!", k, v)
-                else:
-                    gl_check_error("%s '%s' is blacklisted!" % (k, v))
+                log("%s '%s' found in blacklist: %s", k, v, vlist)
+                blacklisted = k, v
+        for k,vlist in WHITELIST.items():
+            v = props.get(k)
+            if v in vlist:
+                log("%s '%s' found in whitelist: %s", k, v, vlist)
+                whitelisted = k, v
+        if blacklisted:
+            if whitelisted:
+                log.info("%s '%s' enabled (found in both blacklist and whitelist)", *whitelisted)
+            elif force_enable:
+                log.warn("Warning: %s '%s' is blacklisted!", *blacklisted)
+            else:
+                gl_check_error("%s '%s' is blacklisted!" % (blacklisted))
 
         #check for specific functions we need:
         from OpenGL.GL import glActiveTexture, glTexSubImage2D, glTexCoord2i, \
@@ -247,9 +300,16 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
         check_functions(glGenProgramsARB, glDeleteProgramsARB, glBindProgramARB, glProgramStringARB)
 
         try:
-            from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE, GL_MAX_TEXTURE_SIZE
+            from OpenGL.GL import GL_MAX_TEXTURE_SIZE
             texture_size = glGetInteger(GL_MAX_TEXTURE_SIZE)
-            rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
+            #this one may be missing?
+            rect_texture_size = texture_size
+            try:
+                from OpenGL.GL import GL_MAX_RECTANGLE_TEXTURE_SIZE
+                rect_texture_size = glGetInteger(GL_MAX_RECTANGLE_TEXTURE_SIZE)
+            except ImportError, e:
+                log("OpenGL: %s", e)
+                log("using GL_MAX_TEXTURE_SIZE=%s as default", texture_size)
         except Exception, e:
             emsg = str(e)
             if hasattr(e, "description"):
@@ -322,16 +382,11 @@ def check_GL_support(gldrawable, glcontext, min_texture_size=0, force_enable=Fal
         gldrawable.gl_end()
 
 def check_support(min_texture_size=0, force_enable=False, check_colormap=False):
-    try:
-        from xpra.platform.paths import get_icon_dir
-        opengl_icon = os.path.join(get_icon_dir(), "opengl.png")
-    except:
-        opengl_icon = None
-    #tricks to get py2exe to include what we need / load it from its unusual path:
-    if sys.platform.startswith("win"):
-        #This is supposed to help py2exe
-        #(must be done after we setup the sys.path in platform.win32.paths):
-        from OpenGL.platform import win32   #@UnusedImport
+    #platform checks:
+    from xpra.platform.gui import gl_check
+    warning = gl_check()
+    if warning:
+        gl_check_error(warning)
 
     props = {}
     import gtk.gdk
@@ -369,14 +424,6 @@ def check_support(min_texture_size=0, force_enable=False, check_colormap=False):
         w = gtk.Window()
         w.set_decorated(False)
         vbox = gtk.VBox()
-        width, height = 32, 32
-        if opengl_icon and os.path.exists(opengl_icon):
-            pixbuf = gtk.gdk.pixbuf_new_from_file(opengl_icon)
-            image = gtk.image_new_from_pixbuf(pixbuf)
-            vbox.add(image)
-            width, height = pixbuf.get_width(), pixbuf.get_height()
-            w.set_default_size(width, height)
-            w.set_resizable(False)
         glarea = gtk.gtkgl.DrawingArea(glconfig)
         glarea.set_size_request(32, 32)
         vbox.add(glarea)
@@ -410,9 +457,11 @@ def check_support(min_texture_size=0, force_enable=False, check_colormap=False):
 
 def main():
     from xpra.platform import init,clean
+    from xpra.platform.gui import init as gui_init
     from xpra.util import pver
     try:
         init("OpenGL-Check")
+        gui_init()
         verbose = "-v" in sys.argv or "--verbose" in sys.argv
         if verbose:
             log.enable_debug()
