@@ -17,8 +17,9 @@ from xpra.log import Logger
 log = Logger("opengl", "paint")
 OPENGL_DEBUG = os.environ.get("XPRA_OPENGL_DEBUG", "0")=="1"
 
+from xpra.os_util import memoryview_to_bytes 
 from xpra.codecs.codec_constants import get_subsampling_divs
-from xpra.client.gl.gl_check import get_DISPLAY_MODE, GL_ALPHA_SUPPORTED
+from xpra.client.gl.gl_check import get_DISPLAY_MODE, GL_ALPHA_SUPPORTED, is_pyopengl_memoryview_safe
 from xpra.client.gl.gl_colorspace_conversions import YUV2RGB_shader, RGBP2RGB_shader
 from xpra.client.gtk2.window_backing import GTK2WindowBacking, fire_paint_callbacks
 from OpenGL import version as OpenGL_version
@@ -99,11 +100,20 @@ if OPENGL_DEBUG:
             glInitFrameTerminatorGREMEDY, glFrameTerminatorGREMEDY)
 from ctypes import c_char_p
 
-
-#support for memory views requires Python 2.7 and PyOpenGL 3.1
-memoryview_type = None
-if sys.version_info[:2]>=(2,7) and OpenGL_version.__version__.split('.')[:2]>=['3','1']:
+try:
+    import OpenGL_accelerate
+except:
+    OpenGL_accelerate = None
+zerocopy_upload = is_pyopengl_memoryview_safe(OpenGL_version.__version__) and OpenGL_accelerate
+try:
     memoryview_type = memoryview
+except:
+    memoryview_type = None
+try:
+    buffer_type = buffer
+except:
+    #not defined in py3k..
+    buffer_type = None
 
 
 # Texture number assignment
@@ -140,6 +150,10 @@ class GLPixmapBacking(GTK2WindowBacking):
     HAS_ALPHA = GL_ALPHA_SUPPORTED
 
     def __init__(self, wid, w, h, window_alpha):
+        #initialize those early as we use them in __repr__
+        self.wid = wid
+        self.size = 0, 0
+        self.pixel_format = None
         alpha = GL_ALPHA_SUPPORTED and window_alpha
         display_mode = get_DISPLAY_MODE(want_alpha=alpha)
         try:
@@ -157,7 +171,11 @@ class GLPixmapBacking(GTK2WindowBacking):
             assert GL_ALPHA_SUPPORTED, "BUG: cannot enable alpha if GL backing does not support it!"
             screen = self._backing.get_screen()
             rgba = screen.get_rgba_colormap()
-            if rgba:
+            display = screen.get_display()
+            if not display.supports_composite():
+                log.warn("display %s does not support compositing, transparency disabled", display.get_name())
+                self._alpha_enabled = False
+            elif rgba:
                 log("%s.__init__() using rgba colormap %s", rgba)
                 self._backing.set_colormap(rgba)
             else:
@@ -182,11 +200,10 @@ class GLPixmapBacking(GTK2WindowBacking):
         self.debug_setup = False
         self.border = None
         self.paint_screen = False
-        self._video_use_swscale = False
         self.draw_needs_refresh = False
         self.offscreen_fbo = None
 
-    def __str__(self):
+    def __repr__(self):
         return "GLPixmapBacking(%s, %s, %s)" % (self.wid, self.size, self.pixel_format)
 
     def init(self, w, h):
@@ -413,9 +430,9 @@ class GLPixmapBacking(GTK2WindowBacking):
             glDisable(GL_TEXTURE_RECTANGLE_ARB)
             #double size since half the line will be off-screen
             glLineWidth(self.border.size*2)
-            glColor4f(self.border.red, self.border.green, self.border.blue, self.border.alpha)
             glBegin(GL_LINE_LOOP)
-            for x,y in ((0, 0), (w, 0), (w, h), (0, h)):
+            glColor4f(self.border.red, self.border.green, self.border.blue, self.border.alpha)
+            for x,y in ((0, 0), (ww, 0), (ww, wh), (0, wh)):
                 glVertex2i(x, y)
             glEnd()
             #reset color to default
@@ -458,12 +475,29 @@ class GLPixmapBacking(GTK2WindowBacking):
             log("%s._do_paint_rgb(..) drawable is not set!", self)
             return False
 
-        #deal with buffers uploads by wrapping them if we can, or copy to a string:
-        if type(img_data)==buffer:
-            if memoryview_type is not None:
-                img_data = memoryview_type(img_data)
+        #prepare the pixel buffer for upload:
+        t = type(img_data)
+        if t==memoryview_type:
+            if not zerocopy_upload:
+                #not safe, make a copy :(
+                img_data = memoryview_to_bytes(img_data)
+                upload = "copy:memoryview_to_bytes"
             else:
+                upload = "zerocopy:memoryview"
+        elif t in (str, buffer_type) and zerocopy_upload:
+            #we can zerocopy if we wrap it:
+            img_data = memoryview_type(img_data)
+            upload = "zerocopy:memoryview", t
+        elif t!=str:
+            if hasattr(img_data, "raw"):
+                img_data = img_data.raw
+                upload = "zerocopy:mmap"
+            else:
+                #everything else.. copy to bytes (aka str):
                 img_data = str(img_data)
+                upload = "copy:str", t
+        else:
+            upload = "copy:str"
 
         try:
             self.set_rgb_paint_state()
@@ -494,7 +528,7 @@ class GLPixmapBacking(GTK2WindowBacking):
             if (rowstride - width * bytes_per_pixel) >= alignment:
                 row_length = width + (rowstride - width * bytes_per_pixel) / bytes_per_pixel
 
-            self.gl_marker("%s %sbpp update at (%d,%d) size %dx%d (%s bytes), stride=%d, row length %d, alignment %d, using GL upload format=%s" % (rgb_format, bpp, x, y, width, height, len(img_data), rowstride, row_length, alignment, CONSTANT_TO_PIXEL_FORMAT.get(pformat)))
+            self.gl_marker("%s %sbpp update at (%d,%d) size %dx%d (%s bytes), stride=%d, row length %d, alignment %d, using GL %s format=%s" % (rgb_format, bpp, x, y, width, height, len(img_data), rowstride, row_length, alignment, upload, CONSTANT_TO_PIXEL_FORMAT.get(pformat)))
 
             # Upload data as temporary RGB texture
             glBindTexture(GL_TEXTURE_RECTANGLE_ARB, self.textures[TEX_RGB])
@@ -527,28 +561,8 @@ class GLPixmapBacking(GTK2WindowBacking):
         return True
 
     def do_video_paint(self, img, x, y, enc_width, enc_height, width, height, options, callbacks):
-        clone = True
-        #we can only use zero copy upload if the image is "thread_safe"
-        #(dec_avcodec2 is not safe...)
-        #and if the data is a memoryview (or if we can wrap it in one)
-        if memoryview_type is not None and img.is_thread_safe():
-            pixels = img.get_pixels()
-            assert len(pixels)==3, "invalid number of planes: %s" % len(pixels)
-            plane_types = list(set([type(plane) for plane in img.get_pixels()]))
-            if len(plane_types)==1:
-                plane_type = plane_types[0]
-                if plane_type==memoryview_type:
-                    clone = False
-                elif plane_type in (str, buffer):
-                    #wrap with a memoryview that pyopengl can use:
-                    views = []
-                    for i in range(3):
-                        views.append(memoryview_type(pixels[i]))
-                    img.set_pixels(views)
-                    clone = False
-        if clone:
-            #copy so the data will be usable (usually a str)
-            img.clone_pixel_data()
+        #copy so the data will be usable (usually a str)
+        img.clone_pixel_data()
         gobject.idle_add(self.gl_paint_planar, img, x, y, enc_width, enc_height, width, height, callbacks)
 
     def gl_paint_planar(self, img, x, y, enc_width, enc_height, width, height, callbacks):
