@@ -16,8 +16,11 @@ import atexit
 import signal
 import socket
 import getpass
+import select
+import time
+import traceback
 
-from xpra.scripts.main import TCP_NODELAY
+from xpra.scripts.main import TCP_NODELAY, warn
 from xpra.dotxpra import DotXpra, ServerSockInUse
 
 o0117 = 79
@@ -25,6 +28,11 @@ o0177 = 127
 o0666 = 438
 o0700 = 448
 
+# use process polling with python versions older than 2.7 and 3.0, (because SIGCHLD support is broken)
+# or when the user requests it with the env var:
+USE_PROCESS_POLLING = os.environ.get("XPRA_USE_PROCESS_POLLING")=="1" or sys.version_info<(2, 7) or sys.version_info[:2]==(3, 0)
+
+
 _cleanups = []
 def run_cleanups():
     global _cleanups
@@ -35,13 +43,20 @@ def run_cleanups():
             c()
         except:
             print("error running cleanup %s" % c)
-            import traceback
             traceback.print_exception(*sys.exc_info())
 
 _when_ready = []
 
+def add_when_ready(f):
+    _when_ready.append(f)
+
+def add_cleanup(f):
+    _cleanups.append(f)
+
+
 def deadly_signal(signum, frame):
-    print("got deadly signal %s, exiting" % {signal.SIGINT:"SIGINT", signal.SIGTERM:"SIGTERM"}.get(signum, signum))
+    sys.stdout.write("got deadly signal %s, exiting\n" % {signal.SIGINT:"SIGINT", signal.SIGTERM:"SIGTERM"}.get(signum, signum))
+    sys.stdout.flush()
     run_cleanups()
     # This works fine in tests, but for some reason if I use it here, then I
     # get bizarre behavior where the signal handler runs, and then I get a
@@ -59,31 +74,22 @@ class ChildReaper(object):
         self._quit = quit_cb
         self._children_pids = {}
         self._dead_pids = set()
+        self._ignored_pids = set()
         from xpra.log import Logger
         self._logger = Logger("server", "util")
-        old_python = sys.version_info < (2, 7) or sys.version_info[:2] == (3, 0)
-        if old_python:
+        if USE_PROCESS_POLLING:
             POLL_DELAY = int(os.environ.get("XPRA_POLL_DELAY", 2))
             self._logger.warn("Warning: outdated/buggy version of Python: %s", ".".join(str(x) for x in sys.version_info))
             self._logger.warn("switching to process polling every %s seconds to support 'exit-with-children'", POLL_DELAY)
-            #keep track of process objects:
-            self.processes = []
-            def check_procs():
-                for proc in self.processes:
-                    if proc.poll() is not None:
-                        self.add_dead_pid(proc.pid)
-                        self.check()
-                self.processes = [proc for proc in self.processes if proc.poll() is None]
-                return True
-            gobject.timeout_add(POLL_DELAY*1000, check_procs)
+            gobject.timeout_add(POLL_DELAY*1000, self.check)
         else:
             #with a less buggy python, we can just check the list of pids
             #whenever we get a SIGCHLD
             #however.. subprocess.Popen will no longer work as expected
             #see: http://bugs.python.org/issue9127
             #so we must ensure certain things that exec happen first:
-            from xpra.version_util import get_platform_info_cache
-            get_platform_info_cache()
+            from xpra.version_util import get_platform_info
+            get_platform_info()
 
             signal.signal(signal.SIGCHLD, self.sigchld)
             # Check once after the mainloop is running, just in case the exit
@@ -94,23 +100,33 @@ class ChildReaper(object):
                 return False # Only call once
             gobject.timeout_add(0, check_once)
 
-    def add_process(self, process, command):
+    def add_process(self, process, command, ignore=False):
         process.command = command
+        assert process.pid>0
         self._children_pids[process.pid] = process
+        if ignore:
+            self._ignored_pids.add(process.pid)
+        self._logger("add_process(%s, %s, %s) pid=%s", process, command, ignore, process.pid)
 
     def check(self):
-        if self._children_pids:
+        pids = set(self._children_pids.keys()) - self._ignored_pids
+        self._logger("check() pids=%s", pids)
+        if pids:
             for pid, proc in self._children_pids.items():
                 if proc.poll() is not None:
                     self.add_dead_pid(pid)
-            pids = set(self._children_pids.keys())
+            self._logger("check() pids=%s, dead_pids=%s", pids, self._dead_pids)
             if pids.issubset(self._dead_pids):
                 self._quit()
+                return False
+        return True
 
     def sigchld(self, signum, frame):
+        self._logger("sigchld(%s, %s)", signum, frame)
         self.reap()
 
     def add_dead_pid(self, pid):
+        self._logger("add_dead_pid(%s)", pid)
         if pid not in self._dead_pids:
             proc = self._children_pids.get(pid)
             if proc:
@@ -124,10 +140,27 @@ class ChildReaper(object):
                 pid, _ = os.waitpid(-1, os.WNOHANG)
             except OSError:
                 break
+            self._logger("reap() waitpid=%s", pid)
             if pid == 0:
                 break
             self.add_dead_pid(pid)
 
+    def get_info(self):
+        d = dict(self._children_pids)
+        info = {"children"          : len(d),
+                "children.dead"     : len(self._dead_pids),
+                "children.ignored"  : len(self._ignored_pids)}
+        i = 0
+        for pid in sorted(d.keys()):
+            proc = d[pid]
+            info["child[%i].live" % i]  = pid not in self._dead_pids
+            info["child[%i].pid" % i]   = pid
+            info["child[%i].command" % i]   = proc.command
+            info["child[%i].ignored" % i] = pid in self._ignored_pids
+            i += 1
+        return info
+
+
 def save_xvfb_pid(pid):
     import gtk
     from xpra.x11.gtk_x11.prop import prop_set
@@ -165,6 +198,10 @@ def xpra_runner_shell_script(xpra_file, starting_dir, socket_dir):
         #XPRA_SOCKET_DIR is a special case, it is handled below
         if var=="XPRA_SOCKET_DIR":
             continue
+        if var.startswith("BASH_FUNC"):
+            #some versions of bash will apparently generate functions
+            #that cannot be reloaded using this script
+            continue
         # :-separated envvars that people might change while their server is
         # going:
         if var in ("PATH", "LD_LIBRARY_PATH", "PYTHONPATH"):
@@ -226,20 +263,20 @@ def write_runner_shell_script(dotxpra, contents, overwrite=True):
 
 
 def display_name_check(display_name):
-    if display_name.startswith(":"):
-        n = display_name[1:]
-        p = n.find(".")
-        if p>0:
-            n = n[:p]
-        try:
-            dno = int(n)
-            if dno>=0 and dno<10:
-                sys.stderr.write("WARNING: low display number: %s\n" % dno)
-                sys.stderr.write("You are attempting to run the xpra server against what seems to be a default X11 display '%s'.\n" % display_name)
-                sys.stderr.write("This is generally not what you want.\n")
-                sys.stderr.write("You should probably use a higher display number just to avoid any confusion (and also this warning message).\n")
-        except:
-            pass
+    """ displays a warning
+        when a low display number is specified """
+    if not display_name.startswith(":"):
+        return
+    n = display_name[1:].split(".")[0]    #ie: ":0.0" -> "0"
+    try:
+        dno = int(n)
+        if dno>=0 and dno<10:
+            sys.stderr.write("WARNING: low display number: %s\n" % dno)
+            sys.stderr.write("You are attempting to run the xpra server against what seems to be a default X11 display '%s'.\n" % display_name)
+            sys.stderr.write("This is generally not what you want.\n")
+            sys.stderr.write("You should probably use a higher display number just to avoid any confusion (and also this warning message).\n")
+    except:
+        pass
 
 
 def get_ssh_port():
@@ -364,22 +401,37 @@ def close_all_fds(exceptions=[]):
             return
     print("Uh-oh, can't close fds, please port me to your system...")
 
-def open_log_file(dotxpra, log_file, display_name):
-    if log_file:
+def shellsub(s, subs={}):
+    """ shell style string substitution using the dictionary given """
+    for var,value in subs.items():
+        s = s.replace("$%s" % var, value)
+        s = s.replace("${%s}" % var, value)
+    return s
+
+def open_log_file(logpath):
+    """ renames the existing log file if it exists,
+        then opens it for writing.
+    """
+    if os.path.exists(logpath):
+        os.rename(logpath, logpath + ".old")
+    return os.open(logpath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, o0666)
+
+def select_log_file(dotxpra, log_file, display_name):
+    """ returns the log file path we should be using given the parameters,
+        this may return a temporary logpath if display_name is not available.
+    """
+    if log_file and display_name:
         if os.path.isabs(log_file):
             logpath = log_file
         else:
             logpath = os.path.join(dotxpra.sockdir(), log_file)
-        logpath = logpath.replace("$DISPLAY", display_name)
-    else:
+        logpath = shellsub(logpath, {"DISPLAY" : display_name})
+    elif display_name:
         logpath = dotxpra.log_path(display_name) + ".log"
-    sys.stderr.write("Entering daemon mode; "
-                     + "any further errors will be reported to:\n"
-                     + ("  %s\n" % logpath))
-    # Do some work up front, so any errors don't get lost.
-    if os.path.exists(logpath):
-        os.rename(logpath, logpath + ".old")
-    return os.open(logpath, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, o0666)
+    else:
+        logpath = os.path.join(dotxpra.sockdir(), "tmp_%d.log" % os.getpid())
+    return logpath
+
 
 def daemonize(logfd):
     os.chdir("/")
@@ -388,17 +440,27 @@ def daemonize(logfd):
     os.setsid()
     if os.fork():
         os._exit(0)
-    close_all_fds(exceptions=[logfd])
+    # save current stdout/stderr to be able to print info
+    # before exiting the non-deamon process
+    # and closing those file descriptors definitively
+    old_fd_stdout = os.dup(1)
+    old_fd_stderr = os.dup(2)
+    close_all_fds(exceptions=[logfd,old_fd_stdout,old_fd_stderr])
     fd0 = os.open("/dev/null", os.O_RDONLY)
     if fd0 != 0:
         os.dup2(fd0, 0)
         os.close(fd0)
+    # reopen STDIO files
+    old_stdout = os.fdopen(old_fd_stdout, "w", 1)
+    old_stderr = os.fdopen(old_fd_stderr, "w", 1)
+    # replace standard stdout/stderr by the log file
     os.dup2(logfd, 1)
     os.dup2(logfd, 2)
     os.close(logfd)
     # Make these line-buffered:
     sys.stdout = os.fdopen(1, "w", 1)
     sys.stderr = os.fdopen(2, "w", 1)
+    return (old_stdout, old_stderr)
 
 
 def sanitize_env():
@@ -418,24 +480,51 @@ def sanitize_env():
              #"XDG_RUNTIME_DIR",
              "QT_GRAPHICSSYSTEM_CHECKED",
              )
-    #force 'simple' / 'xim', as 'ibus' 'immodule' breaks keyboard handling
-    #unless its daemon is also running - and we don't know if it is..
-    #this should override any XSETTINGS too.
+
+def configure_imsettings_env(input_method):
+    im = (input_method or "").lower()
+    if im in ("none", "no"):
+        #the default: set DISABLE_IMSETTINGS=1, fallback to xim
+        #that's because the 'ibus' 'immodule' breaks keyboard handling
+        #unless its daemon is also running - and we don't know if it is..
+        imsettings_env(True, "xim", "xim", "none", "@im=none")
+    elif im=="keep":
+        #do nothing and keep whatever is already set, hoping for the best
+        pass
+    elif im in ("xim", "IBus", "SCIM", "uim"):
+        #ie: (False, "ibus", "ibus", "IBus", "@im=ibus")
+        imsettings_env(True, im.lower(), im.lower(), im, "@im=%s" % im.lower())
+    else:
+        v = imsettings_env(True, im.lower(), im.lower(), im, "@im=%s" % im.lower())
+        warn("using input method settings: %s" % str(v))
+        warn("unknown input method specified: %s" % input_method)
+        warn(" if it is correct, you may want to file a bug to get it recognized")
+
+def imsettings_env(disabled, gtk_im_module, qt_im_module, imsettings_module, xmodifiers):
     #for more information, see imsettings:
     #https://code.google.com/p/imsettings/source/browse/trunk/README
-    os.environ.update({
-               "DISABLE_IMSETTINGS" : "true",
-               "GTK_IM_MODULE"      : "xim",                #or "gtk-im-context-simple"?
-               "QT_IM_MODULE"       : "xim",                #or "simple"?
-               "IMSETTINGS_MODULE"  : "none",               #or "xim"?
-               "XMODIFIERS"         : ""})                  #oe "@im=none"
+    if disabled is True:
+        os.environ["DISABLE_IMSETTINGS"] = "1"                  #this should override any XSETTINGS too
+    elif disabled is False and ("DISABLE_IMSETTINGS" in os.environ):
+        del os.environ["DISABLE_IMSETTINGS"]
+    v = {
+         "GTK_IM_MODULE"      : gtk_im_module,            #or "gtk-im-context-simple"?
+         "QT_IM_MODULE"       : qt_im_module,             #or "simple"?
+         "IMSETTINGS_MODULE"  : imsettings_module,        #or "xim"?
+         "XMODIFIERS"         : xmodifiers,
+         #not really sure what to do with those:
+         #"IMSETTINGS_DISABLE_DESKTOP_CHECK"   : "true",   #
+         #"IMSETTINGS_INTEGRATE_DESKTOP" : "no"}           #we're not a real desktop
+        }
+    os.environ.update(v)
+    return v
 
 def start_pulseaudio(child_reaper, pulseaudio_command):
     from xpra.log import Logger
     log = Logger("sound")
     log("pulseaudio_command=%s", pulseaudio_command)
     pa_proc = subprocess.Popen(pulseaudio_command, stdin=subprocess.PIPE, shell=True, close_fds=True)
-    child_reaper.add_process(pa_proc, "pulseaudio")
+    child_reaper.add_process(pa_proc, "pulseaudio", ignore=True)
     log.info("pulseaudio server started with pid %s", pa_proc.pid)
     def check_pa_start():
         if pa_proc.poll() is not None or pa_proc.pid in child_reaper._dead_pids:
@@ -453,7 +542,6 @@ def start_pulseaudio(child_reaper, pulseaudio_command):
                 log.warn("error trying to stop pulseaudio", exc_info=True)
     _cleanups.append(cleanup_pa)
 
-
 def start_Xvfb(xvfb_str, display_name):
     # We need to set up a new server environment
     xauthority = os.environ.get("XAUTHORITY", os.path.expanduser("~/.Xauthority"))
@@ -463,22 +551,92 @@ def start_Xvfb(xvfb_str, display_name):
         except Exception, e:
             #trying to continue anyway!
             sys.stderr.write("Error trying to create XAUTHORITY file %s: %s\n" % (xauthority, e))
+
+    #identify logfile argument if it exists and if we may have to rename it:
+    xvfb_cmd = xvfb_str.split()
+    if '-logfile' in xvfb_cmd and display_name[0]=='S':
+        xvfb_cmd = xvfb_str.split()
+        logfile_argindex = xvfb_cmd.index('-logfile') + 1
+        assert logfile_argindex<len(xvfb_cmd), "invalid xvfb command string: -logfile should not be last"
+        xorg_log_file = xvfb_cmd[logfile_argindex]
+    else:
+        xorg_log_file = None
+
+    #apply string substitutions:
     subs = {"XAUTHORITY"    : xauthority,
             "USER"          : os.environ.get("USER", "unknown-user"),
             "HOME"          : os.environ.get("HOME", os.getcwd()),
             "DISPLAY"       : display_name}
-    for var,value in subs.items():
-        xvfb_str = xvfb_str.replace("$%s" % var, value)
-        xvfb_str = xvfb_str.replace("${%s}" % var, value)
-    xvfb_cmd = xvfb_str.split()+[display_name]
-    xvfb_executable = xvfb_cmd[0]
-    xvfb_cmd[0] = "%s-for-Xpra-%s" % (xvfb_executable, display_name)
+    xvfb_str = shellsub(xvfb_str, subs)
+
     def setsid():
         #run in a new session
         if os.name=="posix":
             os.setsid()
-    xvfb = subprocess.Popen(xvfb_cmd, executable=xvfb_executable, close_fds=True,
+
+    xvfb_cmd = xvfb_str.split()
+    xvfb_executable = xvfb_cmd[0]
+    if display_name[0]=='S':
+        # 'S' means that we allocate the display automatically
+        r_pipe, w_pipe = os.pipe()
+        xvfb_cmd += ["-displayfd", str(w_pipe)]
+        xvfb_cmd[0] = "%s-for-Xpra-%s" % (xvfb_executable, display_name)
+        def preexec():
+            setsid()
+            #duplicate python's _close_fds() function
+            #(taking care to exclude the pipe)
+            try:
+                MAXFD = os.sysconf("SC_OPEN_MAX")
+            except:
+                MAXFD = 256
+            for i in xrange(3, MAXFD):
+                if i not in (r_pipe, w_pipe):
+                    try:
+                        os.close(i)
+                    except:
+                        pass
+        xvfb = subprocess.Popen(xvfb_cmd, executable=xvfb_executable, close_fds=False,
+                                stdin=subprocess.PIPE, preexec_fn=preexec)
+        # Read the display number from the pipe we gave to Xvfb
+        # waiting up to 10 seconds for it to show up
+        limit = time.time()+10
+        buf = ""
+        while time.time()<limit and len(buf)<8:
+            r, _, _ = select.select([r_pipe], [], [], max(0, limit-time.time()))
+            if r_pipe in r:
+                buf += os.read(r_pipe, 8)
+                if buf[-1] == '\n':
+                    break
+        os.close(r_pipe)
+        os.close(w_pipe)
+        if len(buf) == 0:
+            raise OSError("%s did not provide a display number using -displayfd" % xvfb_executable)
+        if buf[-1] != '\n':
+            raise OSError("%s output not terminated by newline: %s" % (xvfb_executable, buf))
+        try:
+            n = int(buf[:-1])
+        except:
+            raise OSError("%s display number is not a valid number: %s" % (xvfb_executable, buf[:-1]))
+        if n<0 or n>=2**16:
+            raise OSError("%s provided an invalid display number: %s" % (xvfb_executable, n))
+        new_display_name = ":%s" % n
+        sys.stdout.write("Using display number provided by %s: %s\n" % (xvfb_executable, new_display_name))
+        if xorg_log_file != None:
+            #ie: ${HOME}/.xpra/Xorg.${DISPLAY}.log -> /home/antoine/.xpra/Xorg.S14700.log
+            f0 = shellsub(xorg_log_file, subs)
+            subs["DISPLAY"] = new_display_name
+            #ie: ${HOME}/.xpra/Xorg.${DISPLAY}.log -> /home/antoine/.xpra/Xorg.:1.log
+            f1 = shellsub(xorg_log_file, subs)
+            if f0 != f1:
+                os.rename(f0, f1)
+        display_name = new_display_name
+    else:
+        # use display specified
+        xvfb_cmd[0] = "%s-for-Xpra-%s" % (xvfb_executable, display_name)
+        xvfb_cmd.append(display_name)
+        xvfb = subprocess.Popen(xvfb_cmd, executable=xvfb_executable, close_fds=True,
                                 stdin=subprocess.PIPE, preexec_fn=setsid)
+
     from xpra.os_util import get_hex_uuid
     xauth_cmd = ["xauth", "add", display_name, "MIT-MAGIC-COOKIE-1", get_hex_uuid()]
     try:
@@ -488,7 +646,7 @@ def start_Xvfb(xvfb_str, display_name):
     except OSError, e:
         #trying to continue anyway!
         sys.stderr.write("Error running \"%s\": %s\n" % (" ".join(xauth_cmd), e))
-    return xvfb
+    return xvfb, display_name
 
 def check_xvfb_process(xvfb=None):
     if xvfb is None:
@@ -584,7 +742,7 @@ def start_children(child_reaper, commands, fake_xinerama):
             sys.stderr.write("Error spawning child '%s': %s\n" % (child_cmd, e))
 
 
-def run_server(parser, opts, mode, xpra_file, extra_args):
+def run_server(error_cb, opts, mode, xpra_file, extra_args):
     if opts.encoding and opts.encoding=="help":
         #avoid errors and warnings:
         opts.encoding = ""
@@ -622,18 +780,31 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
     #get the display name:
     if shadowing and len(extra_args)==0:
         from xpra.scripts.main import guess_X11_display
-        display_name = guess_X11_display()
+        display_name = guess_X11_display(opts.socket_dir)
+    elif upgrading and len(extra_args)==0:
+        from xpra.scripts.main import guess_xpra_display
+        display_name = guess_xpra_display(opts.socket_dir)
     else:
-        if len(extra_args) != 1:
-            parser.error("need exactly 1 extra argument")
-        display_name = extra_args.pop(0)
-
-    if not shadowing and not proxying:
-        display_name_check(display_name)
+        if len(extra_args) > 1:
+            error_cb("too many extra arguments: only expected a display number")
+        if len(extra_args) == 1:
+            display_name = extra_args[0]
+            if not shadowing and not proxying:
+                display_name_check(display_name)
+        else:
+            if not opts.displayfd:
+                error_cb("displayfd support is not enabled on this system, you must specify the display to use")
+            if opts.use_display:
+                #only use automatic guess for xpra displays and not X11 displays:
+                from xpra.scripts.main import guess_xpra_display     #@Reimport
+                display_name = guess_xpra_display(opts.socket_dir)
+            else:
+                # We will try to find one automaticaly
+                # Use the temporary magic value 'S' as marker:
+                display_name = 'S' + str(os.getpid())
 
     if not shadowing and not proxying and not upgrading and opts.exit_with_children and not opts.start_child:
-        sys.stderr.write("--exit-with-children specified without any children to spawn; exiting immediately")
-        return  1
+        error_cb("--exit-with-children specified without any children to spawn; exiting immediately")
 
     atexit.register(run_cleanups)
     #the server class will usually override those:
@@ -646,15 +817,26 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
     # change if/when we daemonize:
     script = xpra_runner_shell_script(xpra_file, os.getcwd(), opts.socket_dir)
 
+    stdout = sys.stdout
+    stderr = sys.stderr
     # Daemonize:
     if opts.daemon:
         #daemonize will chdir to "/", so try to use an absolute path:
         if opts.password_file:
             opts.password_file = os.path.abspath(opts.password_file)
-
-        logfd = open_log_file(dotxpra, opts.log_file, display_name)
+        # At this point we may not know the display name,
+        # so log_filename0 may point to a temporary file which we will rename later
+        log_filename0 = select_log_file(dotxpra, opts.log_file, display_name)
+        logfd = open_log_file(log_filename0)
         assert logfd > 2
-        daemonize(logfd)
+        stdout, stderr = daemonize(logfd)
+        try:
+            stderr.write("Entering daemon mode; "
+                 + "any further errors will be reported to:\n"
+                 + ("  %s\n" % log_filename0))
+        except:
+            #this can happen if stderr is closed by the caller already
+            pass
 
     # Write out a shell-script so that we can start our proxy in a clean
     # environment:
@@ -663,48 +845,71 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
     from xpra.log import Logger
     log = Logger("server")
 
+    mdns_recs = []
+    sockets = []
     try:
-        # Initialize the sockets before the display,
+        # Initialize the TCP sockets before the display,
         # That way, errors won't make us kill the Xvfb
         # (which may not be ours to kill at that point)
         bind_tcp = parse_bind_tcp(opts.bind_tcp)
-
-        sockets = []
-        mdns_info = {"display" : display_name,
-                     "username": getpass.getuser()}
-        if opts.session_name:
-            mdns_info["session"] = opts.session_name
-        #tcp:
         for host, iport in bind_tcp:
             socket = setup_tcp_socket(host, iport)
             sockets.append(socket)
-        #unix:
-        socket, cleanup_socket = setup_local_socket(dotxpra, display_name, clobber, opts.mmap_group)
-        if socket:      #win32 returns None!
-            sockets.append(socket)
-            if opts.mdns:
-                ssh_port = get_ssh_port()
-                if ssh_port:
-                    mdns_publish(display_name, "ssh", [("", ssh_port)], mdns_info)
         if opts.mdns:
-            mdns_publish(display_name, "tcp", bind_tcp, mdns_info)
+            mdns_recs.append(("tcp", bind_tcp))
     except Exception, e:
         log.error("cannot start server: failed to setup sockets: %s", e)
         return 1
 
     # Do this after writing out the shell script:
-    os.environ["DISPLAY"] = display_name
+    if display_name[0] != 'S':
+        os.environ["DISPLAY"] = display_name
     sanitize_env()
+    configure_imsettings_env(opts.input_method)
 
+    # Start the Xvfb server first to get the display_name if needed
     xvfb = None
     xvfb_pid = None
     if not shadowing and not proxying and not clobber:
         try:
-            xvfb = start_Xvfb(opts.xvfb, display_name)
+            xvfb, display_name = start_Xvfb(opts.xvfb, display_name)
         except OSError, e:
             log.error("Error starting Xvfb: %s\n", e)
             return  1
         xvfb_pid = xvfb.pid
+        #always update as we may now have the "real" display name:
+        os.environ["DISPLAY"] = display_name
+
+    if opts.daemon:
+        log_filename1 = select_log_file(dotxpra, opts.log_file, display_name)
+        if log_filename0 != log_filename1:
+            # we now have the correct log filename, so use it:
+            os.rename(log_filename0, log_filename1)
+            stderr.write("Actual log file name is now: %s\n" % log_filename1)
+        stdout.close()
+        stderr.close()
+
+    if not check_xvfb_process(xvfb):
+        #xvfb problem: exit now
+        return  1
+
+    #setup unix domain socket:
+    socket, cleanup_socket = setup_local_socket(dotxpra, display_name, clobber, opts.mmap_group)
+    if socket:      #win32 returns None!
+        sockets.append(socket)
+        if opts.mdns:
+            ssh_port = get_ssh_port()
+            if ssh_port:
+                mdns_recs.append(("ssh", [("", ssh_port)]))
+
+    #publish mdns records:
+    if opts.mdns:
+        mdns_info = {"display" : display_name,
+                     "username": getpass.getuser()}
+        if opts.session_name:
+            mdns_info["session"] = opts.session_name
+        for mode, listen_on in mdns_recs:
+            mdns_publish(display_name, mode, listen_on, mdns_info)
 
     if not check_xvfb_process(xvfb):
         #xvfb problem: exit now
@@ -764,9 +969,9 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
             return 1
         app = XpraServer()
         app.init(clobber, opts)
-        _cleanups.insert(0, app.cleanup)
-
+    log("%s(%s)", app.init_sockets, sockets)
     app.init_sockets(sockets)
+    log("%s(%s)", app.init_when_ready, _when_ready)
     app.init_when_ready(_when_ready)
 
     #we got this far so the sockets have initialized and
@@ -786,7 +991,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
         def reaper_quit():
             if opts.exit_with_children:
                 log.info("all children have exited and --exit-with-children was specified, exiting")
-                gobject.idle_add(app.clean_quit)
+                gobject.idle_add(app.reaper_quit)
         child_reaper = ChildReaper(reaper_quit)
         if not upgrading and not shadowing and opts.pulseaudio and len(opts.pulseaudio_command)>0:
             start_pulseaudio(child_reaper, opts.pulseaudio_command)
@@ -795,14 +1000,18 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
         if opts.start_child:
             assert os.name=="posix", "start-child cannot be used on %s" % os.name
             start_children(child_reaper, opts.start_child, (opts.fake_xinerama and not shadowing))
+        app.child_reaper = child_reaper
 
     try:
+        log("running %s", app.run)
         e = app.run()
+        log("%s()=%s", app.run, e)
     except KeyboardInterrupt:
+        log.info("stopping on KeyboardInterrupt")
         e = 0
     except Exception, e:
         log.error("server error", exc_info=True)
-        e = 128
+        e = -128
     if e>0:
         # Upgrading/exiting, so leave X server running
         if kill_xvfb in _cleanups:
@@ -814,5 +1023,6 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
             log.info("upgrading: not cleaning up Xvfb or socket")
             # don't delete the new socket (not ours)
             _cleanups.remove(cleanup_socket)
-        log.info("cleanups=%s", _cleanups)
-    return  0
+        log("cleanups=%s", _cleanups)
+        e = 0
+    return e
